/*
  Doc_Arduino - german Arduino Forum
  IDE 1.8.13
  avr-gcc 10.2.0
  09.02.2021
  License: GNU GPLv3
*/

#pragma once

#if defined(ARDUINO_AVR_NANO_EVERY) && defined(__AVR_ATmega4809__)
    #include "include/NanoEveryRegister.h"
#else
    #error "This library only support Arduino Nano Every Board with ATmega4809."
#endif
 
#define INLINE inline __attribute__((always_inline))

namespace NanoEvery 
{ 
  namespace Pin 
  { 
    namespace
    {        
      using Register = volatile uint8_t *;

      #include <assert.h>

      // ------------------ VPORTs ----------------------------------------------------------------------------------------------- //                
      // Direction
      constexpr Register regVPORTdir(const uint8_t pin)       { return (Register) (baseAddr[pin].vport + addrOffset.vDir); }
      // because of pin double assignment 18<>22 and 19<>23
      constexpr Register regVPORTdirPORTA(const uint8_t pin)  { using namespace SpecialPORTA; return (Register) (baseAddrPORTA[pin].vport + addrOffset.vDir); }
      
      // Output Level
      constexpr Register regVPORTout(const uint8_t pin)       { return (Register) (baseAddr[pin].vport + addrOffset.vOut); }
      
      // Input Level Status
      constexpr Register regVPORTin(const uint8_t pin)        { return (Register) (baseAddr[pin].vport + addrOffset.vIn); }
      
      // Interrupt Flag
      constexpr Register regVPORTflag(const uint8_t pin)      { return (Register) (baseAddr[pin].vport + addrOffset.vFlag); }
      
      // ------------------ NORMAL PORTs ----------------------------------------------------------------------------------------- //
      // Pin Mode is Input
      constexpr Register regPORTdirClear(const uint8_t pin)   { return (Register) (baseAddr[pin].port + addrOffset.dirClear); }
                  
      // Input Level Status
      constexpr Register regPORTin(const uint8_t pin)         { return (Register) (baseAddr[pin].port + addrOffset.in); }
      
      // Pin Mode is Output
      constexpr Register regPORTdirSet(const uint8_t pin)     { return (Register) (baseAddr[pin].port + addrOffset.dirSet); }
      
      // Pin Level High
      constexpr Register regPORToutSet(const uint8_t pin)     { return (Register) (baseAddr[pin].port + addrOffset.outSet); }
      
      // Pin Level Low
      constexpr Register regPORToutClear(const uint8_t pin)   { return (Register) (baseAddr[pin].port + addrOffset.outClear); }
      
      // Pin Level Toggle
      constexpr Register regPORToutToggle(const uint8_t pin)  { return (Register) (baseAddr[pin].port + addrOffset.outToggle); }
      
      // Interrupt Flag
      constexpr Register regPORTflag(const uint8_t pin)       { return (Register) (baseAddr[pin].port + addrOffset.flag); }
      
      // ------------------ SPECIALs ------------------ //
      // SlewRate
      constexpr Register regPortCtrl(const uint8_t pin)       { return (Register) (baseAddr[pin].port + addrOffset.portCtrl); }
      
      // Invert, Input-Level, Pullup, Interrupt
      constexpr Register regPinCtrl(const uint8_t pin)       { return (Register) (baseAddr[pin].port + baseAddr[pin].pinCtrl); }
      // because of pin double assignment 18<>22 and 19<>23
      constexpr Register regPinCtrlPORTA(const uint8_t pin)  { using namespace SpecialPORTA; return (Register) (baseAddrPORTA[pin].port + baseAddrPORTA[pin].pinCtrl); }
      
      // Pin Bitmaske
      constexpr uint8_t getMask(const uint8_t pin)            { return (baseAddr[pin].mask); } 
      // because of pin double assignment 18<>22 and 19<>23
      constexpr uint8_t getMaskPORTA(const uint8_t pin)       { using namespace SpecialPORTA; return (baseAddrPORTA[pin].mask); } 
    }

    /*
    hier in Abhängigkeit von DEBUG NanoEveryPinDegging.h inkludieren und damit zusätzliche Methoden zur Verfügung stellen
    */
    
    template<uint8_t arduinoPin>
    class Pin 
    {         
      static_assert(arduinoPin < 22, "Pin number too large, valid range 0...21");
 
      constexpr void specialTreatmentI2Cpins (void)
      {
        if ((arduinoPin == 18) || (arduinoPin == 19))
        { // because of pin double assignment
          TWI0_MCTRLA = 0;   // I2C Master Mode disable
          TWI0_SCTRLA = 0;   // I2C Slave Mode disable
        }
        if (arduinoPin == 18)
        { // PA2 switch to Input, prevents that the PF2 and PA2 are outputs at the same time
          // PF2 is the standard I/O Pin, PA2 is used for I2C
          *regPinCtrlPORTA(0) = *regPinCtrlPORTA(0) & ~PORT_PULLUPEN_bm;
          *regVPORTdir(0) = *regVPORTdir(0) & ~getMaskPORTA(0);
        }
        if (arduinoPin == 19)
        { // PA3 switch to Input, prevents that the PF3 and PA3 are outputs at the same time
          // PF3 is the standard I/O Pin, PA3 is used for I2C
          *regPinCtrlPORTA(1) = *regPinCtrlPORTA(1) & ~PORT_PULLUPEN_bm;
          *regVPORTdir(1) = *regVPORTdir(1) & ~getMaskPORTA(1);
        }
      }
     
      protected:
        
        // Inputs
        void INLINE initInput() {
          specialTreatmentI2Cpins();
          *regVPORTdir(arduinoPin) = *regVPORTdir(arduinoPin) & ~getMask(arduinoPin);
        }
        
        bool INLINE isOn() const {
          return *regVPORTin(arduinoPin) & getMask(arduinoPin);
        }
    
        void INLINE enablePullup() { 
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) | PORT_PULLUPEN_bm;      // Bit 3 Pullup
        }
        
        void INLINE disablePullup() { 
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) & ~PORT_PULLUPEN_bm;    
        }
                   
        // Outputs
        void INLINE initOutput() {
          specialTreatmentI2Cpins();
          *regVPORTdir(arduinoPin) = *regVPORTdir(arduinoPin) | getMask(arduinoPin);
        }
        
        void INLINE setOn() {
          *regVPORTout(arduinoPin) = *regVPORTout(arduinoPin) | getMask(arduinoPin);
        }
        
        void INLINE setOff() {
          *regVPORTout(arduinoPin) = *regVPORTout(arduinoPin) & ~getMask(arduinoPin);
        }
        
        void INLINE toggle() {
          *regVPORTin(arduinoPin) = *regVPORTin(arduinoPin) | getMask(arduinoPin);
        }
        
        // Specials
        void INLINE enableSlewRate() { 
          *regPortCtrl(arduinoPin) = *regPortCtrl(arduinoPin) | PORT_SRL_bm;        // Bit 0 SlewRate, valid for the complete port
        }
        
        void INLINE disableSlewRate() { 
          *regPortCtrl(arduinoPin) = *regPortCtrl(arduinoPin) & ~PORT_SRL_bm;       
        }
        
        void INLINE enableInvertIO() { 
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) | PORT_INVEN_bm;       // Bit 7 Inverted I/O
        }
        
        void INLINE disableInvertIO() {  
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) & ~PORT_INVEN_bm;       
        }
              
        // Interrupts
        void INLINE reset() { 
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) & ~PORT_ISC_gm;        // all Interrupts disabled but input buffer enabled
        }
        
        void INLINE senseBothEdges() { 
          reset();      
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) | 0x01;                // Interrupt enabled with sense on both edges
        }
        
        void INLINE senseRising() { 
          reset();       
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) | 0x02;                // Interrupt enabled with sense on rising edge
        }
        
        void INLINE senseFalling() { 
          reset();      
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) | 0x03;                // Interrupt enabled with sense on falling edge
        }
        
        void INLINE senseDisable() { 
          reset();        
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) | 0x04;                // Interrupt and digital input buffer disabled
        }
        
        void INLINE senseLow() { 
          reset();        
          *regPinCtrl(arduinoPin) = *regPinCtrl(arduinoPin) | 0x05;                // Interrupt enabled with sense on low level
        }     

        void INLINE deleteFlag() {       
          *regVPORTflag(arduinoPin) = getMask(arduinoPin);                           // delete Interrupt Flag
        }
    };

    template<uint8_t arduinoPin>
    class InputPin : protected Pin<arduinoPin> 
    {
      public:      
        using Pin<arduinoPin>::enablePullup;
        using Pin<arduinoPin>::disablePullup;
        using Pin<arduinoPin>::isOn;
        using Pin<arduinoPin>::enableInvertIO;
        using Pin<arduinoPin>::disableInvertIO;
        
        void  INLINE init()
        {
          Pin<arduinoPin>::initInput();
        }
      
        INLINE operator bool() const
        {
          return Pin<arduinoPin>::isOn();
        }
    };
    
    template<uint8_t arduinoPin>
    class InterruptPin : protected InputPin<arduinoPin> 
    {
      public:        
        using Pin<arduinoPin>::reset;
        using Pin<arduinoPin>::enablePullup;
        using Pin<arduinoPin>::disablePullup;
        using Pin<arduinoPin>::senseBothEdges;
        using Pin<arduinoPin>::senseRising;
        using Pin<arduinoPin>::senseFalling;
        using Pin<arduinoPin>::senseLow;
        using Pin<arduinoPin>::senseDisable;
        using Pin<arduinoPin>::deleteFlag;
        using Pin<arduinoPin>::isOn;
              
        void INLINE init()
        {
          Pin<arduinoPin>::initInput();
        }          
    };
    
    template<uint8_t arduinoPin>
    class TasterGND : protected InputPin<arduinoPin>
    {
      public:
        using InputPin<arduinoPin>::init;
        using Pin<arduinoPin>::enablePullup;
        //using Pin<arduinoPin>::disablePullup;
        //using Pin<arduinoPin>::enableInvertIO;
        //using Pin<arduinoPin>::disableInvertIO;
        
        void  INLINE init() {
          Pin<arduinoPin>::enableInvertIO();
          Pin<arduinoPin>::initInput();
        }
        
        bool INLINE pressed() const     { return InputPin<arduinoPin>::isOn(); }
        
        bool INLINE released() const    { return !InputPin<arduinoPin>::isOn(); }
        
        INLINE operator bool() const    { return pressed(); }
        
        bool INLINE operator ()() const { return pressed(); }
    };
        
    template<uint8_t arduinoPin>
    class OutputPin : protected Pin<arduinoPin>
    {
      public:       
        using Pin<arduinoPin>::setOn;
        using Pin<arduinoPin>::setOff;
        using Pin<arduinoPin>::toggle;
        using Pin<arduinoPin>::isOn;
        using Pin<arduinoPin>::enableSlewRate;
        using Pin<arduinoPin>::disableSlewRate;
        using Pin<arduinoPin>::enableInvertIO;
        using Pin<arduinoPin>::disableInvertIO;
      
        void INLINE init() { Pin<arduinoPin>::initOutput(); }
       
        void INLINE set(const bool value) {
          if(value) setOn();
          else setOff();
        }
        
        bool INLINE operator = (const bool value) {
          set(value);
          return value;
        }
        
        bool INLINE operator () (const bool value) {
          set(value);
          return value;
        }
        
        INLINE operator bool() const { return Pin<arduinoPin>::isOn(); }
    };

    template<uint8_t arduinoPin>
    class OutputPinInvert : public OutputPin<arduinoPin>
    {
      public:
        void INLINE init()
        { 
          Pin<arduinoPin>::initOutput();
          Pin<arduinoPin>::enableInvertIO();
        }

        INLINE operator bool() const { return !Pin<arduinoPin>::isOn(); }
    };
  } 
}
